# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel

from .customer import Customer
from .mailbox_ref import MailboxRef


class Rating(BaseModel):

    customer = properties.Instance(
        'Partial customer object.',
        instance_class=Customer,
        required=True,
    )
    ticket_id = properties.Integer(
        'Ticket ID',
        required=True,
    )
    thread_id = properties.Integer(
        'Thread ID',
        required=True,
    )
    mailbox = properties.Instance(
        'Reference to the mailbox that the conversation belongs to.',
        instance_class=MailboxRef,
        required=True,
    )
    rating = properties.StringChoice(
        'Satisfaction rating.',
        choices=['Great', 'Okay', 'Bad'],
        required=True,
    )
    comments = properties.String(
        'Additional comments',
    )
    created_at = properties.DateTime(
        'UTC time when this rating was created.',
    )
    modified_at = properties.DateTime(
        'UTC time when this rating was modified.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Phone(BaseModel):

    value = properties.String(
        'Phone Number',
        required=True,
    )
    location = properties.StringChoice(
        'Location for this phone number.',
        choices=['home',
                 'fax',
                 'mobile',
                 'pager',
                 'work',
                 'other',
                 ],
        default='other',
        required=True,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .. import BaseApi

from ..models.customer import Customer

from ..request_paginator import RequestPaginator


class Customers(BaseApi):
    """This represents the ``Customers`` Endpoint.

    The following aspects are implemented:
    
    * `List Customers
      <http://developer.helpscout.net/help-desk-api/customers/list/>`_
    * `Get Customer
      <http://developer.helpscout.net/help-desk-api/customers/get/>`_
    * `Create Customer
      <http://developer.helpscout.net/help-desk-api/customers/create/>`_
    * `Update Customer
      <http://developer.helpscout.net/help-desk-api/customers/update/>`_
    """

    __object__ = Customer

    @classmethod
    def create(cls, session, customer, _reload=False):
        """Create a customer.

        Args:
            session (requests.Session): Authenticated requests Session.
            customer (Customer): The customer to be created.
            _reload (bool, optional): Set this request parameter to ``True``
            to return the created conversation in the response.

        Returns:
            bool: Success status, if ``_reload`` is ``False``.
            Customer: Newly created customer, if ``_reload`` is ``True``.
        """
        data = {
            'customer': customer.to_api(),
            'reload': _reload,
        }
        return cls(
            '/customers.json',
            data=data,
            request_type=RequestPaginator.POST,
            singleton=True,
            session=session,
        )

    @classmethod
    def get(cls, session, customer_id):
        """Return a specific customer.

        Args:
            session (requests.Session): Authenticated requests Session.
            customer_id (int): The ID of the customer to get.

        Returns:
            Customer: A customer singleton, if existing.
            None: If no match for the ID.
        """
        return cls(
            '/customers/%d.json' % customer_id,
            singleton=True,
            session=session,
        )

    @classmethod
    def list(cls, session, first_name=None, last_name=None, email=None,
             modified_since=None):
        """List the customers.

        Customers can be filtered on any combination of first name, last name,
        email, and modifiedSince.

        Args:
            session (requests.Session): Authenticated requests Session.
            first_name (str, optional): First name of customer.
            last_name (str, optional): Last name of customer.
            email (str, optional): Email address of customer.
            modified_since (datetime, optional): If modified after this date.

        Returns:
            RequestPaginator of Customer: Customers iterator.
        """
        data = cls.__object__.get_non_empty_vals({
            'firstName': first_name,
            'lastName': last_name,
            'email': email,
            'modifiedSince': modified_since,
        })
        return cls('/customers.json', data, session=session)

    @classmethod
    def update(cls, session, customer, _reload=False):
        """Update a customer.

        Args:
            session (requests.Session): Authenticated requests Session.
            customer (Customer): The customer to be updated.
            _reload (bool, optional): Set this request parameter to ``True``
            to return the created conversation in the response.

        Returns:
            bool: Success status, if ``_reload`` is ``False``.
            Customer: Freshly updated customer, if ``_reload`` is ``True``.
        """
        data = {
            'customer': customer,
            'reload': _reload,
        }
        return cls(
            '/customers/%d.json' % customer.id,
            data=data,
            request_type=RequestPaginator.PUT,
            singleton=True,
            session=session,
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .request_paginator import RequestPaginator

from .base_model import BaseModel


class BaseApi(object):
    """This is the API interface object to be implemented by API adapters.
    
    It acts as a collection for the API object that it represents, passing
    through iteration to the API's request paginator.

    Attributes:
        BASE_URI (str): HelpScout API URI base.
        paginator (RequestPaginator): Object to use for producing an iterator
         representing multiple requests (API response pages). Created on init.
        __object__ (helpscout.models.BaseModel): Model object that API
         represents.
    """

    BASE_URI = 'https://api.helpscout.net/v1'

    # This should be replaced in child classes with the correct model.
    __object__ = BaseModel

    # This is set within new, after the object has been created.
    paginator = None

    def __new__(cls, endpoint, data=None,
                request_type=RequestPaginator.GET, singleton=False,
                session=None):
        """Create a new API object.
        
        Args:
            endpoint (str): The API endpoint that this represents.
            ``BASE_URI`` will be prepended, with no slashes added.
            data (dict, optional): Data to send with the request.
            request_type (str, optional): Type of request (``GET or ``POST``).
            Defaults to ``GET``.
            singleton (bool, optional): Set this to ``True`` to assert that
            there is not more than one result, and return the first result (or
            ``None`` if there is no result.
            session (requests.Session, optional): An authenticated requests
            session to use.
        
        Raises:
            HelpScoutRemoteException: If ``singleton`` is ``True``, but the
            remote API responds with more than one result.

        Returns:
            BaseApi: An instance of an API object, if ``singleton`` is
            ``False``.
            BaseModel: An instance of a Model, if ``singleton`` is
            ``True`` and there are results.
            None: If ``singleton`` is ``True`` and there are no results.
        """
        paginator = RequestPaginator(
            endpoint='%s%s' % (cls.BASE_URI, endpoint),
            data=data,
            output_type=cls.__object__.from_api,
            request_type=request_type,
            session=session,
        )
        if singleton:
            results = paginator.call(paginator.data)
            result_length = len(results)
            if result_length == 0:
                return None
            return results[0]
        obj = super(BaseApi, cls).__new__(cls)
        obj.paginator = paginator
        return obj

    def __iter__(self):
        """Pass through iteration to the API response."""
        for row in self.paginator:
            yield row
        raise StopIteration()



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel

from .email import Email
from .mailbox_ref import MailboxRef
from .person import Person
from .source import Source
from .tag import Tag


class Conversation(BaseModel):
    """This represents a conversation."""

    type = properties.StringChoice(
        'The type of conversation.',
        choices=['email', 'chat', 'phone'],
        required=True,
    )
    folder_id = properties.Integer(
        'ID of the Mailbox Folder to which this conversation resides.',
        required=True,
    )
    is_draft = properties.Bool(
        'Is this a draft conversation?',
    )
    number = properties.Integer(
        'The conversation number displayed in the UI. This number can be used '
        'in combination with the id to construct a URL to the conversation on '
        'the Help Scout website. Example: '
        '``https://secure.helpscout.net/conversation/<id>/<number>/``',
        required=True,
    )
    owner = properties.Instance(
        'The Help Scout user who is currently assigned to this conversation.',
        instance_class=Person,
        required=True,
    )
    mailbox = properties.Instance(
        'The mailbox to which this conversation belongs.',
        instance_class=MailboxRef,
        required=True,
    )
    customer = properties.Instance(
        'The customer who this conversation is associated with.',
        instance_class=Person,
        required=True,
    )
    thread_count = properties.Integer(
        'This count represents the number of published threads found on the '
        'conversation (it does not include line items, drafts or threads held '
        'for review by Traffic Cop).',
        required=True,
    )
    status = properties.StringChoice(
        'Status of the conversation.',
        choices=['action', 'pending', 'closed', 'spam'],
        default='pending',
        required=True,
    )
    subject = properties.String(
        'Subject of conversation.',
    )
    preview = properties.String(
        'Conversation preview.',
    )
    created_by = properties.Instance(
        'The ``Person`` who created this conversation. The ``type`` property '
        'will specify whether it was created by a ``user`` or a ``customer``.',
        instance_class=Person,
    )
    created_at = properties.DateTime(
        'UTC time when this conversation was created.',
    )
    modified_at = properties.DateTime(
        'UTC time when a user last modified this conversation.',
    )
    closed_at = properties.DateTime(
        'UTC time when this conversation was closed. Null if not closed.',
    )
    closed_by = properties.Instance(
        'The Help Scout user who closed this conversation.',
        instance_class=Person,
    )
    source = properties.Instance(
        'Specifies the method in which this conversation was created.',
        instance_class=Source,
    )
    cc = properties.List(
        'Emails that are CCd.',
        prop=Email,
    )
    bcc = properties.List(
        'Emails that are BCCd.',
        prop=Email,
    )
    tags = properties.List(
        'Tags for the conversation',
        prop=Tag,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import unittest

from .. import AuthProxy
from .. import HelpScout
from ..apis import __all__ as all_apis


class TestHelpScout(unittest.TestCase):

    API_KEY = 'test key'

    def setUp(self):
        super(TestHelpScout, self).setUp()
        self.hs = HelpScout(self.API_KEY)

    def test_init_session(self):
        """It should build a session with the proper authentication."""
        self.assertEqual(self.hs.session.auth.username, self.API_KEY)

    def test_load_apis(self):
        """It should load all available APIs."""
        self.assertEqual(len(self.hs.__apis__), len(all_apis))

    def test_api_instance_attributes(self):
        """It should properly set all of the API instance attributes."""
        for api_name, api in self.hs.__apis__.items():
            self.assertEqual(getattr(self.hs, api_name), api)

    def test_api_auth_proxy(self):
        """It should wrap the APIs in an AuthProxy."""
        for api in self.hs.__apis__.values():
            self.assertIsInstance(api, AuthProxy)



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class SocialProfile(BaseModel):

    value = properties.String(
        'Username',
        required=True,
    )
    location = properties.StringChoice(
        'Type of social profile.',
        choices=['twitter',
                 'facebook',
                 'linkedin',
                 'aboutme',
                 'google',
                 'googleplus',
                 'tungleme',
                 'quora',
                 'foursquare',
                 'youtube',
                 'flickr',
                 'other',
                 ],
        default='other',
        required=True,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import requests

from .exceptions import HelpScoutRemoteException, HelpScoutValidationError


class RequestPaginator(object):
    """ RequestPaginator provides an iterator based upon an initial request.
    """

    # Response attributes that mean things
    PAGE_TOTAL = 'pages'  # Total number of pages
    PAGE_CURRENT = 'page'  # Current page number
    PAGE_DATA_MULTI = 'items'  # Attribute if multiple results
    PAGE_DATA_SINGLE = 'item'  # Attribute if one result

    SSL_VERIFY = True  # Verify SSL
    PAGE_SIZE = 50  # Page size returned by HelpScout

    # HTTP operation constants
    DELETE = 'delete'
    GET = 'get'
    POST = 'post'
    PUT = 'put'

    # Starting page ints
    page_current = 0
    page_total = 0

    def __init__(self, endpoint, data=None, output_type=dict,
                 request_type=GET, session=None):
        """Initialize the RequestPaginator object.

        Args:
            endpoint (str): URI endpoint to call.
            session (requests.Session): The authenticated requests session.
            data (dict): Data to be sent in the query string for the
             Request.
            output_type (type): Class type to output. Object will be
             instantiated using the current row before output.
            request_type (str): Type of request to send (``GET`` or ``POST``).
            session (requests.Session, optional): An authenticated requests
             session to use.

        Raises:
            NotImplementedError: In the event that an invalid request type was
             defined.
        """
        self.endpoint = endpoint
        self.data = data
        self.output_type = output_type
        if request_type not in (self.GET, self.POST, self.PUT, self.DELETE):
            raise NotImplementedError(
                'The `%s` request type is not implemented', request_type,
            )
        self.request_type = request_type
        self.session = session or requests.Session()

    def __iter__(self, page=1):
        """Provide an iterator for the remote request.

        The result is returned as an instantiated `self.output_type`.
        """
        self.page_current = page
        data = self.data.copy()
        result = self.call(data)
        for row in result:
            yield self.output_type(**row)
        if self.page_current < self.page_total:
            for inner_row in self.__iter__(self.page_current + 1):
                yield inner_row
        else:
            raise StopIteration()

    def call(self, data=None):
        """Generic API caller. Return the JSON decoded result.

        Args:
            data (dict, optional): Either the request parameters or the JSON
             data, depending on the request type.

        Raises:
            NotImplementedError: In the event that an invalid request type was
             defined.

        Returns:
            mixed: JSON decoded respons.
        """
        return getattr(self, self.request_type)(data)

    def delete(self, json=None):
        """Send a DELETE request and return the JSON decoded result.

        Args:
            json (dict, optional): Object to encode and send in request.

        Returns:
            mixed: JSON decoded response data.
        """
        return self._call('delete', url=self.endpoint, json=json)

    def get(self, params=None):
        """Send a POST request and return the JSON decoded result.
        
        Args:
            params (dict, optional): Mapping of parameters to send in request.
        
        Returns:
            mixed: JSON decoded response data.
        """
        return self._call('get', url=self.endpoint, params=params)

    def post(self, json=None):
        """Send a POST request and return the JSON decoded result.

        Args:
            json (dict, optional): Object to encode and send in request.

        Returns:
            mixed: JSON decoded response data.
        """
        return self._call('post', url=self.endpoint, json=json)

    def put(self, json=None):
        """Send a PUT request and return the JSON decoded result.

        Args:
            json (dict, optional): Object to encode and send in request.

        Returns:
            mixed: JSON decoded response data.
        """
        return self._call('put', url=self.endpoint, json=json)

    def _call(self, method, *args, **kwargs):
        """Call the remote service and return the response data."""

        assert self.session
        method = getattr(self.session, method)

        if not kwargs.get('verify'):
            kwargs['verify'] = self.SSL_VERIFY

        response = method(*args, **kwargs)
        response_json = response.json()

        if response.status_code < 200 or response.status_code >= 300:
            message = response_json.get('error', response_json.get('message'))
            raise HelpScoutRemoteException(response.status_code, message)

        self.page_current = response_json[self.PAGE_CURRENT]
        self.page_total = response_json[self.PAGE_TOTAL]

        try:
            return response_json[self.PAGE_DATA_MULTI]
        except KeyError:
            pass

        try:
            return [response_json[self.PAGE_DATA_SINGLE]]
        except KeyError:
            pass

        return True





# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .request_paginator import RequestPaginator

from .exceptions import HelpScoutRemoteException


class AuthProxy(object):
    """This object acts as a transparent authentication proxy for the API.

    This is required because the API objects are naive of the authentication
    that is setup in ``HelpScout``, and all of the API interface methods are
    ``classmethods`` because they are instantiating.
    """

    # These methods will not pass through to ``proxy_class``.
    METHOD_NO_PROXY = [
        'auth_proxy',
    ]

    def __init__(self, session, proxy_class):
        """Instantiate an API Authentication Proxy.

        Args:
            auth (requests.Session): Authenticated requests Session.
            proxy_class (type): A class implementing the ``BaseApi``
             interface.
        """
        assert isinstance(proxy_class, type)
        self.session = session
        self.proxy_class = proxy_class

    def __getattr__(self, item):
        """Override attribute getter to act as a proxy for``proxy_class``.

        If ``item`` is contained in ``METHOD_NO_PROXY``, it will not be
        proxied to the ``proxy_class`` and will instead return the attribute
        on this object.

        Args:
            item (str): Name of attribute to get.
        """
        if item in self.METHOD_NO_PROXY:
            return super(AuthProxy, self).__getattr__(item)
        attr = getattr(self.proxy_class, item)
        if callable(attr):
            return self.auth_proxy(attr)

    def auth_proxy(self, method):
        """Authentication proxy for API requests.

        This is required because the API objects are naive of ``HelpScout``,
        so they would otherwise be unauthenticated.

        Args:
            method (callable): A method call that should be authenticated. It
             should accept a ``requests.Session`` as its first parameter,
             which should be used for the actual API call.

        Returns:
            mixed: The results of the authenticated callable.
        """
        def _proxy(*args, **kwargs):
            """The actual proxy, which instantiates and authenticates the API.

            Args:
                *args (mixed): Args to send to class instantiation.
                **kwargs (mixed): Kwargs to send to class instantiation.

            Returns:
                mixed: The result of the authenticated callable.
            """
            return method(self.session, *args, **kwargs)

        return _proxy



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties
import re

from .exceptions import HelpScoutValidationError


# Identify lowerCamelCase strings.
# https://stackoverflow.com/a/1176023/861399
REGEX_CAMEL_FIRST = re.compile(r'(.)([A-Z][a-z]+)')
REGEX_CAMEL_SECOND = re.compile(r'([a-z0-9])([A-Z])')


class BaseModel(properties.HasProperties):
    """This is the model that all other models inherit from.

    It provides some convenience functions, and the standard ``id`` property.
    """

    id = properties.Integer(
        'Unique identifier',
    )

    @classmethod
    def from_api(cls, **kwargs):
        """Create a new instance from API arguments.

        This will switch camelCase keys into snake_case for instantiation.

        It will also identify any ``Instance`` or ``List`` properties, and
        instantiate the proper objects using the values. The end result being
        a fully Objectified and Pythonified API response.

        Raises:
            HelpScoutValidationError: In the event that an unexpected property
             is received.

        Returns:
            BaseModel: Instantiated model using the API values.
        """

        vals = cls.get_non_empty_vals({
            cls._to_snake_case(k): v for k, v in kwargs.items()
        })

        for attribute, value in vals.items():

            prop = cls._props.get(attribute)

            if not prop:
                raise HelpScoutValidationError(
                    '"%s" is not a valid property for "%s".' % (
                        attribute, cls,
                    ),
                )

            if isinstance(prop, properties.Instance):
                vals[attribute] = prop.instance_class.from_api(**value)

            elif isinstance(prop, properties.List):
                vals[attribute] = [
                    prop.prop.instance_class.from_api(**v) for v in value
                ]

        return cls(**vals)

    def to_api(self):
        """Return a dictionary to send to the API.
        
        Returns:
            dict: Mapping representing this object that can be sent to the
             API.
        """
        vals = {}
        for attribute, attribute_type in self._props.items():

            prop = getattr(self, attribute)

            if isinstance(attribute_type, properties.Instance):
                    prop = prop.to_api()

            elif isinstance(attribute_type, properties.List):
                prop = [p.to_api() for p in prop]

            vals[self._to_camel_case(attribute)] = prop

        return vals

    @staticmethod
    def get_non_empty_vals(mapping):
        """Return the mapping without any ``None`` values."""
        return {
            k: v for k, v in mapping.items() if v is not None
        }

    @staticmethod
    def _to_snake_case(string):
        sub_string = r'\1_\2'
        string = REGEX_CAMEL_FIRST.sub(sub_string, string)
        return REGEX_CAMEL_SECOND.sub(sub_string, string).lower()

    @staticmethod
    def _to_camel_case(string):
        components = string.split('_')
        return '%s%s' % (
            components[0],
            ''.join(c.title() for c in components[1:]),
        )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

from .conversations import Conversations
from .customers import Customers
from .mailboxes import Mailboxes
from .teams import Teams
from .users import Users


__all__ = [
    'Conversations',
    'Customers',
    'Mailboxes',
    'Teams',
    'Users',
]



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class AttachmentData(BaseModel):

    data = properties.String(
        'base64 encoded data.',
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class MailboxRef(BaseModel):
    """The mailbox ref is a subset of the full Mailbox object."""

    name = properties.String(
        'Name of the Mailbox.',
        required=True,
    )



# -*- coding: utf-8 -*-
# Copyright 2017-TODAY LasLabs Inc.
# License MIT (https://opensource.org/licenses/MIT).

import properties

from .. import BaseModel


class Address(BaseModel):
    """This represents an address."""

    lines = properties.List(
        'Address line strings',
        prop=properties.String(
            'Address line string',
        ),
    )
    city = properties.String(
        'City',
        required=True,
    )
    state = properties.String(
        'State',
        required=True,
    )
    postal_code = properties.String(
        'Postal Code',
        required=True,
    )
    country = properties.String(
        'Country',
        required=True,
    )
    created_at = properties.DateTime(
        'UTC time when this address was created.',
    )
    modified_at = properties.DateTime(
        'UTC time when this address was modified.',
    )
